/* 2008 (c) Intel Corporation,  All rights reserved.
 *
 * Original author Øyvind Kolås <pippin@openedhand.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     - Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     - Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     - Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Intel Corporation ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL Intel Corporation BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdarg.h>
#include <clutter/clutter.h>
#include <clutter-box2d.h>

#include <string.h>

#include "moblin-desktop.h"
#include "moblin-disc.h"
#include "mock-friends.h"     

#define APP_ZOOMED_OUT   -1000.0

ClutterActor *painter_new ();
ClutterActor *test_new ();
ClutterActor *settings_new ();
ClutterActor *fifteen_new ();
ClutterActor *table_new (const gchar *path);

void moblin_swap_out ();
void friends_start (void);

static MoblinDesktop _desktop;

MoblinDesktop *moblin_desktop(void)
{
  return &_desktop;
}
#define DEFAULT_Z_CAMERA    0.869f


#define TEX_COORD "gl_TexCoord[0]"
#define COLOR_VAR "gl_Color"

#define FRAGMENT_SHADER_VARS                                    \
  "uniform sampler2D tex;"                                      \
  "uniform float x_step, y_step;"                               \

#define FRAGMENT_SHADER_BEGIN                                   \
  "void main (){"                                               \
  "  vec4 color = texture2D (tex, vec2(" TEX_COORD "));"

#define FRAGMENT_SHADER_END                    \
      "  gl_FragColor = color;"    \
      "  gl_FragColor = gl_FragColor * " COLOR_VAR ";" \
      "}"

static const gchar *glsl_blur =
     FRAGMENT_SHADER_VARS
     "uniform float radius;"
     FRAGMENT_SHADER_BEGIN
     "float u, v;"
     "float v0 =" TEX_COORD ".t - radius * y_step;"
     "float v1 ="TEX_COORD ".t + radius * y_step;"
     "int count = 4;" /* the initial count of 4 makes us make it darker */
     "for (u=" TEX_COORD ".s - radius * x_step; u < "TEX_COORD ".s + radius * x_step; u += x_step)"
     "  for (v=v0; v<v1; v += y_step)"
     "    {"
     "      color += texture2D(tex, vec2(u,v));"
     "      count ++;"
     "    }"
     "color = color / float(count);"
     /*"color = texture2D(tex, vec2(gl_TexCoord[0].s + x_step * 20,"
     "                            gl_TexCoord[0].t + y_step * 20));"*/
     FRAGMENT_SHADER_END;



void moblin_blur (void)
{
  MoblinDesktop *desktop = moblin_desktop ();
  ClutterShader *shader;
  GError        *error = NULL;
 
  g_assert (!desktop->maincopy);
  if (desktop->maincopy)
    return;

#if 0
  desktop->maincopy = clutter_texture_new_from_actor (desktop->mainstuff);
#else
  {
    guchar *pixels;
    gint    i;
    gint    w,h;
    ClutterTexture *texture;

    w = clutter_actor_get_width (clutter_stage_get_default ());
    h = clutter_actor_get_height (clutter_stage_get_default ());

    pixels = clutter_stage_read_pixels (CLUTTER_STAGE (desktop->stage),
                                        0, 0, w, h);

    /* force the pixels opaque */
    for (i=0;i< w*h;i++)
      pixels[i*4+3]=255;

  /* FIXME: make this work properly for rotated desktops */

    texture = CLUTTER_TEXTURE (clutter_texture_new ());
    clutter_actor_set_size (CLUTTER_ACTOR (texture), w, h);
    clutter_texture_set_from_rgb_data (texture, pixels, TRUE, w, h, w*4,4, 0, NULL);
    desktop->maincopy = CLUTTER_ACTOR (texture);
  }
#endif

  clutter_group_add (desktop->copyholder, desktop->maincopy);
  clutter_actor_hide (desktop->mainstuff);

  shader = clutter_shader_new ();
  clutter_shader_set_fragment_source (shader, glsl_blur, -1);
  clutter_shader_compile (shader, &error);

  if (error)
    {
      g_print ("unable to compile shader : %s\n", error->message);
      g_error_free (error);
    }

  clutter_actor_set_shader (desktop->maincopy, shader);
  {
    gint tex_width = clutter_actor_get_width (desktop->maincopy);
    gint tex_height = clutter_actor_get_height (desktop->maincopy);
    tex_width = clutter_util_next_p2 (tex_width);
    tex_height = clutter_util_next_p2 (tex_height);

    /* store the needed stepping in normalized coordinates */
    clutter_actor_set_shader_param (desktop->maincopy, "x_step",
                                    1.0f / tex_width);
    clutter_actor_set_shader_param (desktop->maincopy, "y_step",
                                    1.0f / moblin_desktop_width ());
  }
 clutter_actor_set_shader_param (desktop->maincopy, "radius", 3.0);
}

void moblin_unblur (void)
{
  MoblinDesktop *desktop = moblin_desktop ();
  g_assert (desktop->maincopy);
  clutter_actor_show (desktop->mainstuff);
  clutter_actor_destroy (desktop->maincopy);
  desktop->maincopy = NULL;
}


gdouble
depth_to_scale (gint depth)
{
  ClutterActor *stage = clutter_stage_get_default ();

  gfloat        z_camera = DEFAULT_Z_CAMERA;

  /* FIXME: should handle fov != 60 */

  z_camera *= clutter_actor_get_width (stage);
  return (z_camera / (z_camera - depth));
}

gint
scale_to_depth (gdouble scale)
{
  ClutterActor *stage = clutter_stage_get_default ();

  gfloat        z_camera = DEFAULT_Z_CAMERA;

  /* FIXME: should handle fov != 60 */
  z_camera *= clutter_actor_get_width (stage);

  return (z_camera * scale - z_camera) / scale;
}

static void
desktop_key_release_cb (ClutterStage           *stage,
                        ClutterKeyEvent        *kev,
                        gpointer                user_data)
{
  switch (clutter_key_event_symbol (kev))
    {
      case CLUTTER_Escape:
        clutter_main_quit ();
        break;
      case CLUTTER_Tab:
        moblin_fullscreen (!moblin_desktop()->wheels);
        break;
      case CLUTTER_q:
        moblin_swap_out ();
        break;
      case CLUTTER_a:
        moblin_desktop_activate_application (test_new ());
        break;
      case CLUTTER_s:
        moblin_desktop_activate_application (settings_new (NULL));
        break;
    }
}

static gboolean
launcher_activated (ClutterActor *launcher,
                    ClutterEvent *event,
                    gpointer      callback)
{
  MoblinDesktop *desktop = moblin_desktop ();
  ClutterActor * (*launch_cb)(gpointer data) = callback;
  ClutterActor *app;
  app = launch_cb(NULL);

  clutter_actor_set_opacity (app, 0);
  clutter_actor_set_position (app, moblin_desktop_width () * 2,
                                   moblin_desktop_height () - clutter_actor_get_height(app)*
                                   desktop->wheel_child_factor
   );
  clutter_actor_set_scale (app, desktop->wheel_child_factor, desktop->wheel_child_factor);
  moblin_desktop_activate_application (app);
  return TRUE;
}


void
moblin_desktop_init (gint    argc,
                     gchar **argv)
{
  ClutterActor *stage;
  ClutterColor  stage_color = { 0x00, 0x00, 0x00, 0x00 };
  MoblinDesktop *desktop = moblin_desktop ();

  stage = clutter_stage_get_default ();
  clutter_stage_set_color (CLUTTER_STAGE (stage), &stage_color);

  desktop->width2 = 800;
  desktop->height2 = 480;

  {
    gchar **arg = argv + 1;
    while (*arg)
      {
       if (g_str_equal (*arg, "-h") ||
          g_str_equal (*arg, "--help"))
        {
          g_print ("There is no help\n");
          return;
        }
      else if (g_str_equal (*arg, "-s"))
        {
          arg++; g_assert (*arg);
          desktop->width2 = atoi (*arg);
          if (strstr (*arg, "x"))
            desktop->height2 = atoi (strstr (*arg, "x") + 1);
        }
      else if (g_str_equal (*arg, "-fs"))
        {
          desktop->fullscreen = TRUE;
        }
      else if (g_str_equal (*arg, "-r"))
        {
          desktop->rotated = TRUE;
        }
      else
        {
        }
      arg++;
      }
  }
  clutter_actor_set_size (stage, desktop->width2, desktop->height2);


  if (desktop->rotated)
    {
      clutter_actor_set_rotation (stage, CLUTTER_Z_AXIS, 90.0,
                                  desktop->width2/2, desktop->width2/2,0);
    }

  clutter_actor_show (stage);
  desktop->stage = stage;
  desktop->background = clutter_texture_new_from_file (ASSET_ROOT "background.png",
                                                       NULL);
  if (desktop->background == NULL)
    {
      desktop->background = clutter_rectangle_new ();
    }

  {
    gdouble       inv_scale = 20;
    gdouble       scale     = 1.0 / inv_scale;
    clutter_actor_set_scale (desktop->background, inv_scale, inv_scale);

    clutter_actor_set_position (desktop->background, -inv_scale * moblin_desktop_width () / 2 + (moblin_desktop_width ()/2),
                                -inv_scale * moblin_desktop_height () / 2 + (moblin_desktop_height () / 2));
    clutter_actor_set_depth (desktop->background, scale_to_depth (scale));
  }



  desktop->wheel_radius = moblin_desktop_height ()/2;

  desktop->middle = clutter_group_new ();

  desktop->left = g_object_new (MOBLIN_TYPE_DISC,
                             "radius", desktop->wheel_radius,
                             "slots", 9,
                             NULL);
  {
    gint i;
    gchar **dir_entries = get_dir_entries (ASSET_ROOT "tasks");
    gint    count = pasp_count (dir_entries);

    for (i=0; i< 0; i++)
      {
         ClutterActor *child;
         child = clutter_texture_new_from_file (dir_entries[g_random_int_range(0,count)], NULL);

         clutter_container_add (CLUTTER_CONTAINER (desktop->left), child, NULL);
      }
    g_free (dir_entries);
  }



  desktop->right = g_object_new (MOBLIN_TYPE_DISC,
                             "radius", desktop->wheel_radius,
                             "slots", 9,
                             NULL);
  {
    gint i;
    gchar **dir_entries = get_dir_entries (ASSET_ROOT "apps");
    gint    count = pasp_count (dir_entries);

    for (i=0; i<9 && i<count; i++)
      {
         ClutterActor *child;
         child = g_object_new (CLUTTER_TYPE_TEXTURE,
             "filename", dir_entries[i],
             "filter-quality", CLUTTER_TEXTURE_QUALITY_HIGH,
             NULL);
               
         clutter_container_add (CLUTTER_CONTAINER (desktop->right), child, NULL);

         { /* scale the icons so that they end up being 0.2 the height of the display */
           gint  height = clutter_actor_get_height (child);
           gfloat factor =    (moblin_desktop_height () * 0.32) / height;

           desktop->wheel_child_factor = factor * 0.2;
           clutter_actor_set_scale (child, factor, factor);
         }
         clutter_actor_set_reactive (child, TRUE);

         if (strstr (dir_entries[i], "camera"))
           {
             g_signal_connect (child, "button-press-event",
                               G_CALLBACK (launcher_activated), table_new);
           }
         else if (strstr (dir_entries[i], "settings"))
           {
             g_signal_connect (child, "button-press-event",
                               G_CALLBACK (launcher_activated), settings_new);
           }
         else if (strstr (dir_entries[i], "game"))
           {
             g_signal_connect (child, "button-press-event",
                               G_CALLBACK (launcher_activated), fifteen_new);
           }
         else
           {
             g_signal_connect (child, "button-press-event",
                               G_CALLBACK (launcher_activated), test_new);
           }
      }
    g_free (dir_entries);
  }

#if 1
  clutter_actor_set_rotation (desktop->left,  CLUTTER_Y_AXIS,  20, 0, 0, 0);
  clutter_actor_set_rotation (desktop->right, CLUTTER_Y_AXIS, -20, 0, 0, 0);
#endif

  desktop->mainstuff = clutter_group_new ();
  desktop->copyholder = clutter_group_new ();
  desktop->modal     = clutter_group_new ();


  clutter_container_add (CLUTTER_CONTAINER (stage),
                         desktop->mainstuff,
                         desktop->copyholder,
                         desktop->modal,
                         NULL);

  clutter_container_add (CLUTTER_CONTAINER (desktop->mainstuff),
                         desktop->background,
                         desktop->middle,
                         desktop->left,
                         desktop->right,
                         NULL);

  clutter_actor_set_position (desktop->left,
                              -1.25 * desktop->wheel_radius, 
			      moblin_desktop_height () * 0.5);

  clutter_actor_set_position (desktop->right,
      moblin_desktop_width ()  + 1.25 * desktop->wheel_radius,
      moblin_desktop_height () * 0.5);

  moblin_fullscreen (FALSE);

  g_signal_connect (stage,
                    "key-release-event",
                    G_CALLBACK (desktop_key_release_cb),
                    NULL);

  clutter_actor_raise_top (desktop->left);
  clutter_actor_raise_top (desktop->right);

}

void moblin_fullscreen (gboolean state)
{
  MoblinDesktop *desktop = moblin_desktop ();
  if (state)
    {
      tween (desktop->left, 300,
                  "x", -0.25 * desktop->wheel_radius,
                  "y", moblin_desktop_height ()*0.5,
                  NULL);
      tween (desktop->right, 300,
                  "x", moblin_desktop_width () +
		  0.25 * desktop->wheel_radius,
                  "y", moblin_desktop_height ()*0.5,
                  NULL);

      tween (desktop->middle, 300,
                  "depth", APP_ZOOMED_OUT,
                  NULL);
    }
  else
    {
      tween (desktop->left, 300,
                  "x", desktop->wheel_radius * -2.0,
                  "y", moblin_desktop_height ()*0.5,
	      NULL);

      tween (desktop->right, 300,
                  "x", moblin_desktop_width ()*1.0 + desktop->wheel_radius*2,
                  "y", moblin_desktop_height ()*0.5,
                  NULL);

      tween (desktop->middle, 300,
                  "depth", 0.0,
                  NULL);
    }
  desktop->wheels = state;


}

void copy_geometry (ClutterActor *from,
                    ClutterActor *to)
{
  gint x, y;
  gdouble rot;
  gdouble scale;
  gint depth;

  g_object_get (from, "x", &x,
                      "y", &y,
                      "depth", &depth,
                      "rotation-angle-z", &rot,
                      "scale-x", &scale,
                      NULL);
  g_object_set (to,   "x", x,
                      "y", y,
                      "depth", depth,
                      "rotation-angle-z", rot,
                      "scale-x", scale,
                      "scale-y", scale,
                      NULL);
}

static gboolean final_swap (gpointer data)
{
  ClutterActor *shuttle = data;
  ClutterActor *app;
  gdouble scale;
  GList *children;
  children = clutter_container_get_children (CLUTTER_CONTAINER (shuttle));
  if (!children)
    return FALSE;
  g_assert (children);
  g_assert (children->next == NULL);
  app = children->data;
  g_list_free (children);

  if (!app)
    return FALSE;

  clutter_actor_get_scale (shuttle, &scale, NULL);
  clutter_actor_set_scale (app, scale, scale);
  clutter_actor_reparent (app, moblin_desktop()->left);

  clutter_actor_destroy (shuttle);
  return FALSE;
}

static void
moblin_actor_get_transformed_position (ClutterActor *actor,
                                       gint         *x,
                                       gint         *y)
{
   clutter_actor_get_transformed_position (actor, x, y);

   /* this might be a bug ? */
   /* XXX: ugly little hack */
   if (moblin_desktop()->rotated)
     {
       gint f;
       f=*x;
       *x=*y;*y=f;
       *y = moblin_desktop_height () - *y;
     }
}

static gboolean task_in_disc_capture (ClutterActor *app,
                                      ClutterEvent *event,
                                      gpointer      data)
{
  MoblinDesktop *desktop = moblin_desktop ();
  switch (clutter_event_type (event))
    {
      case CLUTTER_BUTTON_PRESS:
          {
            gint x, y;
            ClutterActor *box = clutter_actor_get_parent (app);
            gdouble middle_scale;
            g_object_ref (app);

        middle_scale = depth_to_scale (clutter_actor_get_depth (desktop->middle));

#if 1
        moblin_actor_get_transformed_position (app, &x, &y);

#endif



        x -= moblin_desktop_width ()/2;
        x /= middle_scale;
        x += moblin_desktop_width ()/2;

        y -= moblin_desktop_height ()/2;
        y /= middle_scale;
        y += moblin_desktop_height ()/2;


        clutter_container_remove_actor (CLUTTER_CONTAINER (box), app);
        clutter_actor_set_position (app, x, y);
        clutter_actor_set_depth (app, 0);
        clutter_actor_set_scale (app, desktop->wheel_child_factor, desktop->wheel_child_factor);

        /* turn of tracking by passing NULL as other */
        clutter_box2d_actor_track (app, NULL, CLUTTER_BOX2D_TRACK_POSITION);
        moblin_desktop_activate_application (app);

        g_signal_handlers_disconnect_by_func(app, task_in_disc_capture, data);
          }
        return FALSE;
        break;
      default:
        break;
    }
  return TRUE;
}

static void attach_reactivate_on_press (ClutterActor *actor)
{
  clutter_actor_set_reactive (actor, TRUE);
  g_signal_connect (actor, "captured-event",
                           G_CALLBACK (task_in_disc_capture), actor);
}


static gboolean moblin_desktop_focused_swap_out (gpointer data)
{
  ClutterActor *shuttle;
  ClutterActor *old_app;
  GList *children;

  /* retrieve the running application from the desktop */
  MoblinDesktop *desktop = moblin_desktop ();
  children = clutter_container_get_children (CLUTTER_CONTAINER (desktop->middle));
  if (!children)
    return FALSE;
  g_assert (children);
  g_assert (children->next == NULL);
  old_app = children->data;
  g_list_free (children);

  attach_reactivate_on_press (old_app);

  shuttle = clutter_group_new ();
  clutter_group_add (CLUTTER_GROUP(desktop->mainstuff), shuttle);
  copy_geometry (desktop->middle, shuttle);
  clutter_actor_reparent (old_app, shuttle);

  {
    gint    x, y;
    gdouble scale = desktop->wheel_child_factor;
    moblin_disc_get_next_xy (MOBLIN_DISC (desktop->left), &x, &y);

    x += clutter_actor_get_x (desktop->left);
    y += clutter_actor_get_y (desktop->left);
    x -= clutter_actor_get_width (old_app) * scale * 0.5;
    y -= clutter_actor_get_height (old_app) * scale * 0.5;

    tween (shuttle, 300,
                "x",     x*1.0,
                "y",     y*1.0,
                "scale", scale,
                "depth", 0.0,
                NULL);
    g_timeout_add (400, final_swap, shuttle);
  }
  return FALSE;
}

/* moves the current centered app to the side */
void moblin_swap_out ()
{
  moblin_fullscreen (TRUE);
  moblin_desktop_focused_swap_out (NULL);
}


gboolean launcher_task (gpointer data)
{
  ClutterActor  *actor = data;
  gboolean       has_children = FALSE;
  MoblinDesktop *desktop = moblin_desktop ();
  GList         *children;

  children = clutter_container_get_children (CLUTTER_CONTAINER (desktop->middle));
  if (children)
    {
      has_children = TRUE;
      g_list_free (children);
    }

  if (has_children)
    {
      /* return to main context once, will try to launch again later */
      return TRUE;
    }

    if (clutter_actor_get_parent (actor) != desktop->middle)
      {
        clutter_container_add (CLUTTER_CONTAINER (desktop->middle), actor, NULL);
      }
    /*clutter_actor_set_opacity (actor, 0);*/
    tween (actor, 500, 
     
        "x", (moblin_desktop_width () - clutter_actor_get_width (actor)) * 0.5,
        "y", (moblin_desktop_height () - clutter_actor_get_height (actor)) * 0.5,
        "scale", 1.0,
        "depth", 0.0,
        "rotation", 0.0,
        "opacity", 255.0, NULL);
    return FALSE;
}

void
moblin_desktop_activate_application (ClutterActor *actor)
{
  moblin_desktop_focused_swap_out (NULL);
  if (launcher_task (actor))
    {
      g_timeout_add (120, launcher_task, actor);
    }
}




typedef struct {
  gchar        *string;
  GCallback     callback;
  gpointer      userdata;
  ClutterActor *actor;
} MoblinNotification;


static gboolean
notification_press (ClutterActor *label,
                    ClutterEvent *event,
                    gpointer      data)
{
  return FALSE;
}

static gboolean
notification_release (ClutterActor *label,
                      ClutterEvent *event,
                      gpointer      data)
{
  MoblinNotification *notification = data;
  void (*callback) (const gchar *string, gpointer userdata);
  callback = (void*)notification->callback;

  if (callback)
    callback (notification->string, notification->userdata);
  return FALSE;
}

void moblin_add_notification (const gchar *string,
                              GCallback    callback,
                              gpointer     userdata)
{
  MoblinDesktop      *desktop = moblin_desktop ();
  MoblinNotification *notification;

#if 0
  if (!strstr (string, "☎"))
    return; /* we filter non calls for now it is too busy otherwize */
#endif

  if (desktop->notification_group == NULL)
  {
    /* create the notification group and background the first time around */
    desktop->notification_group = clutter_group_new ();
    desktop->notification_bg = clutter_texture_new_from_file (ASSET_ROOT "notification.png", NULL);

    clutter_container_add (CLUTTER_CONTAINER (desktop->mainstuff), desktop->notification_group, NULL);
    clutter_container_add (CLUTTER_CONTAINER (desktop->notification_group),
                           desktop->notification_bg, NULL);

    clutter_actor_set_position (desktop->notification_group,
       (moblin_desktop_width () - clutter_actor_get_width (desktop->notification_bg))/2,
	      
       (moblin_desktop_height () - clutter_actor_get_height (desktop->notification_bg)));
    clutter_actor_set_opacity (desktop->notification_group, 0);
  }

  /* Make sure the notification is visible... */
  tween (desktop->notification_group, 500, "opacity", 255.0, NULL);

  notification = g_new0 (MoblinNotification, 1);
  notification->string = g_strdup (string);
  notification->callback = callback;
  notification->userdata = userdata;
  notification->actor = clutter_label_new_full ("Arial 20px", string,
                                  &(ClutterColor){0xff, 0xff, 0xff, 0xff});
  desktop->notifications = g_list_append (desktop->notifications, notification);

  clutter_group_add (CLUTTER_GROUP (desktop->notification_group),
                     notification->actor);

  clutter_actor_set_reactive (notification->actor, TRUE);
  g_signal_connect (notification->actor, "button-press-event",
                    G_CALLBACK (notification_press), notification);
  g_signal_connect (notification->actor, "button-release-event",
                    G_CALLBACK (notification_release), notification);

  clutter_actor_set_position(
	  notification->actor,
	  40, clutter_actor_get_height (desktop->notification_bg));

  {
    GList *iter;
    for (iter = desktop->notifications; iter; iter = iter->next)
      {
        MoblinNotification *notification = iter->data;
        if (!notification->actor)
          continue;
        gint y = clutter_actor_get_y (notification->actor);
	if (y == clutter_actor_get_height (desktop->notification_bg)) 
	{
		/* tween this one into the middle */
		tween(
			notification->actor, 300,
			"y",  
			clutter_actor_get_height(desktop->notification_bg) / 
			2.0 - 12.0,
			"opacity", 255.0,
			NULL);
	} else if (y == 0) {
		/* delete this one... */
		clutter_actor_destroy (notification->actor);
		notification->actor = NULL; /* marked for deletion handled in loop below */
	} else {
		/* tween this one away... */
		tween(
			notification->actor, 300,
			"y", 0.0,
			"opacity", 0.0,
			NULL);
          }
      }

    /* XXX: quirky way to clean up with a goto, but it keeps the above loop
     * simpler
     */
    again:
    for (iter = desktop->notifications; iter; iter = iter->next)
      {
        MoblinNotification *notification = iter->data;
        if (notification->actor == NULL)
          {
            desktop->notifications = g_list_remove (desktop->notifications, notification);
            g_free (notification);
            goto again;
          }
      }
  }
}

gint
moblin_desktop_width (void)
{
  if (moblin_desktop()->rotated)
    return moblin_desktop ()->height2;
  return moblin_desktop ()->width2;
}

gint
moblin_desktop_height (void)
{
  if (moblin_desktop()->rotated)
    return moblin_desktop ()->width2;
  return moblin_desktop ()->height2;
}



gint
main (int   argc,
      char *argv[])
{
  ClutterActor *stage;

  clutter_init (&argc, &argv);

  stage = clutter_stage_get_default ();
  moblin_desktop_init (argc, argv);

  if(0)moblin_desktop_activate_application (table_new (ASSET_ROOT "photos/"));
  if(1)moblin_desktop_activate_application (test_new ());

  mock_friends ();

  /*
  g_object_set (custom_cursor (1000,1000,0), "hot-x", 110, "hot-y", 110,
                                       "normal", clutter_texture_new_from_file (ASSET_ROOT "cursor.png", NULL),
                                       "pressed", clutter_texture_new_from_file (ASSET_ROOT "cursor-pressed.png", NULL),
                                       NULL);
                                       */

  clutter_main ();

  return EXIT_SUCCESS;
}
