/* 2008 (c) Intel Corporation,  All rights reserved.
 *
 * Original author Øyvind Kolås <pippin@openedhand.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     - Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     - Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     - Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Intel Corporation ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL Intel Corporation BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <clutter/clutter.h>
#include <cogl/cogl.h>
#include "clutter-box2d.h"
#include <string.h>
#include <math.h>
#include "moblin-disc.h"
#include "util.h"

#define MAX_SLOTS 64

G_DEFINE_TYPE (MoblinDisc, moblin_disc, CLUTTER_TYPE_BOX2D);

enum
{
  PROP_0,
  PROP_RADIUS,
  PROP_SLOTS,
};

#define MOBLIN_DISC_GET_PRIVATE(obj) \
   (G_TYPE_INSTANCE_GET_PRIVATE ((obj), MOBLIN_TYPE_DISC, MoblinDiscPrivate))

struct _MoblinDiscPrivate
{
  gint radius;
  gint slots;
  ClutterActor *knobs[MAX_SLOTS];
  ClutterActor *items[MAX_SLOTS];
};

static void
moblin_disc_set_property (GObject      *object,
                          guint         prop_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
  MoblinDisc        *cursor;
  MoblinDiscPrivate *priv;

  cursor = MOBLIN_DISC(object);
  priv = cursor->priv;

  switch (prop_id) 
    {
    case PROP_RADIUS:
      priv->radius = g_value_get_int (value);
      break;
    case PROP_SLOTS:
      priv->slots = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
moblin_disc_get_property (GObject    *object,
                      guint       prop_id,
                      GValue     *value,
                      GParamSpec *pspec)
{
  MoblinDiscPrivate *priv = MOBLIN_DISC (object)->priv;

  switch (prop_id) 
    {
    case PROP_RADIUS:
      g_value_set_int (value, priv->radius);
      break;
    case PROP_SLOTS:
      g_value_set_int (value, priv->slots);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    } 
}

static gint moblin_disc_next_unused (MoblinDisc *disc)
{
  gint i;
  MoblinDiscPrivate *priv;
  priv = disc->priv;

  for (i=0; i< priv->slots; i++)
    if (priv->items[i]==NULL)
      return i;
  return -1;
}



static void moblin_disc_actor_added   (MoblinDisc       *disc,
                                   ClutterActor *actor,
                                   gpointer      user_data)
{
  MoblinDiscPrivate *priv;
  ClutterActor  *knob;

  gint slot = moblin_disc_next_unused (disc);
  gint x, y;
  gdouble scale;
  priv = disc->priv;

  priv->items[slot] = actor;
  knob = priv->knobs[slot];

  g_assert (knob);
  g_assert (actor);
  x = clutter_actor_get_x (knob);
  y = clutter_actor_get_y (knob);

  clutter_actor_get_scale (actor, &scale, NULL);
  clutter_actor_set_position (actor,
                              x - scale*clutter_actor_get_width (actor)/2, 
                              y - scale*clutter_actor_get_height (actor)/2);

  /* lock the relative position between child and knob 
   */
  clutter_box2d_actor_track (actor, knob, CLUTTER_BOX2D_TRACK_POSITION);
}


static void moblin_disc_actor_removed (MoblinDisc       *disc,
                                   ClutterActor *actor,
                                   gpointer      user_data)
{
  gint i;
  MoblinDiscPrivate *priv;
  priv = disc->priv;

  for (i=0; i< priv->slots; i++)
    if (priv->items[i]==actor)
      priv->items[i]=NULL;
}


void moblin_disc_get_next_xy (MoblinDisc *disc, gint *x, gint *y)
{
  MoblinDiscPrivate *priv;
  ClutterActor  *knob;
  priv = MOBLIN_DISC (disc)->priv;
  gint slot = moblin_disc_next_unused (disc);
  knob = priv->knobs[slot];

  g_assert (knob);
  clutter_actor_get_position (knob, x, y);
}

static GObject *
moblin_disc_constructor (GType                  type,
                     guint                  n_params,
                     GObjectConstructParam *params)

{
  GObject      *object;
  ClutterActor *box2d;
  ClutterActor *center = clutter_rectangle_new ();
  ClutterActor *wheel = clutter_rectangle_new ();
  ClutterColor  transparent = {0xff, 0xff, 0xff, 0x00};
  MoblinDiscPrivate *priv;

  object = G_OBJECT_CLASS (moblin_disc_parent_class)->constructor (
                                                        type, n_params, params);
  box2d  = CLUTTER_ACTOR (object);
  priv   = MOBLIN_DISC (object)->priv;

  clutter_actor_set_size (center, 16, 16);
  clutter_actor_set_size (wheel, priv->radius*2, priv->radius*2);

  clutter_actor_set_position (center, -8, -8);
  clutter_actor_set_position (wheel, -priv->radius, -priv->radius);

  clutter_rectangle_set_color (CLUTTER_RECTANGLE (center), &transparent);

  /* we cannot hide the wheel, it needs to be shown to be pickable,
   * thus we make it transparent instead
   */
  clutter_rectangle_set_color (CLUTTER_RECTANGLE (wheel), &transparent);

  clutter_actor_hide (center);

  clutter_container_add (CLUTTER_CONTAINER (box2d), wheel, center, NULL);

  clutter_container_child_set (CLUTTER_CONTAINER (box2d), center,
                               "mode", CLUTTER_BOX2D_STATIC,
                               NULL);
  clutter_container_child_set (CLUTTER_CONTAINER (box2d), wheel,
                               "mode", CLUTTER_BOX2D_DYNAMIC,
                               "manipulatable", TRUE,
                               NULL);

  /* pin the center to a static axle using a revolute joint 
   */
  {
    ClutterVertex anchor1 = { CLUTTER_UNITS_FROM_FLOAT (8.0),
                             CLUTTER_UNITS_FROM_FLOAT (8.0) };
    ClutterVertex anchor2 = { CLUTTER_UNITS_FROM_FLOAT (priv->radius),
                             CLUTTER_UNITS_FROM_FLOAT (priv->radius) };
    clutter_box2d_add_revolute_joint (CLUTTER_BOX2D (box2d),
        center, wheel,
        &anchor1, &anchor2, 0.0);
  }

   {
    gint i;

    /*
     * create virtual knobs where children can be "hung"
     */

    for (i=0; i< priv->slots; i++)
      {
        gint x = cos ((i-priv->slots*0.25) * 3.1415 * 2 / priv->slots) * priv->radius * 0.88;
        gint y = sin ((i-priv->slots*0.25) * 3.1415 * 2 / priv->slots) * priv->radius * 0.88;

        ClutterActor *knob;
       
        ClutterVertex anchor1 = { CLUTTER_UNITS_FROM_FLOAT (x + priv->radius),
                                  CLUTTER_UNITS_FROM_FLOAT (y + priv->radius)};
        ClutterVertex anchor2 = { CLUTTER_UNITS_FROM_FLOAT (1.0),
                                  CLUTTER_UNITS_FROM_FLOAT (0.0) };

        knob = clutter_rectangle_new ();
        clutter_rectangle_set_color (CLUTTER_RECTANGLE (knob), &transparent);
        clutter_actor_set_size (knob, 2, 2);

        anchor2.x = CLUTTER_UNITS_FROM_FLOAT (0);

        clutter_actor_set_position (knob, x, y);
        clutter_container_add (CLUTTER_CONTAINER (box2d), knob, NULL);

        clutter_container_child_set (CLUTTER_CONTAINER (box2d), knob,
                                     "mode", CLUTTER_BOX2D_DYNAMIC, NULL);

        clutter_box2d_add_revolute_joint (CLUTTER_BOX2D (box2d),
            wheel, knob,
            &anchor1, &anchor2, 0.0);

        priv->knobs[i]=knob;
      }
   }


  g_signal_connect_after (object, "actor-added",
                          G_CALLBACK (moblin_disc_actor_added), NULL);
  clutter_box2d_set_simulating (CLUTTER_BOX2D (object), TRUE);
  g_signal_connect (object, "actor-removed",
                    G_CALLBACK (moblin_disc_actor_removed), NULL);


  return object;
}



static void
moblin_disc_class_init (MoblinDiscClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  /*ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);*/

  gobject_class->constructor  = moblin_disc_constructor;
  gobject_class->set_property = moblin_disc_set_property;
  gobject_class->get_property = moblin_disc_get_property;

  g_type_class_add_private (gobject_class, sizeof (MoblinDiscPrivate));

#define PARAM_FLAGS (G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |    \
                     G_PARAM_STATIC_BLURB | G_PARAM_CONSTRUCT_ONLY |\
                     G_PARAM_READABLE | G_PARAM_WRITABLE)

  g_object_class_install_property (gobject_class,
                                   PROP_RADIUS,
                                   g_param_spec_int (
                                     "radius",
                                     "Radius-Y",
                                     "radius of disc contact points",
                                     0, G_MAXINT,
                                     0,
                                     PARAM_FLAGS));

  g_object_class_install_property (gobject_class,
                                   PROP_SLOTS,
                                   g_param_spec_int (
                                     "slots",
                                     "Slots",
                                     "number of slots for children on wheel",
                                     0, G_MAXINT,
                                     0,
                                     PARAM_FLAGS));


#undef PARAM_FLAGS
}

static void
moblin_disc_init (MoblinDisc *self)
{
  MoblinDiscPrivate *priv;

  self->priv = priv = MOBLIN_DISC_GET_PRIVATE (self);

  priv->radius = 111;
}
