/* 2008 (c) OpenedHand, authored by Øyvind Kolås */

/* This file contains a small set of mock contacts/friends, each friend
 * is a state machine that probabilistically goes from being away to being
 * awake (online), then perhaps chatty from which stage it might send call
 * requests or instant messages.
 */

#include <glib.h>
#include <glib/gprintf.h>
#include <clutter/clutter.h>
#include "moblin-desktop.h"
#include "mock-friends.h"

typedef enum {
  AWAY = 0,
  AWAKE,
  CHATTY,
  CALLING,
  TALKING,
  PRODDING,
  CHATTING
} FriendState;

const gchar *state_strings[]=
{
  "away",
  "awake",
  "chatty",
  "calling",
  "talking",
  "prodding",
  "chatting",
  NULL
};

typedef struct Friend
{
  const gchar *name;
  gint         state;
} Friend;

static Friend friends[]=
{
  {"Frank Smith", CHATTY},
  {"Sarah Lee", AWAKE},
  {"Mom", CHATTY},
  {NULL, }
};

static void callback (const gchar *string, 
                      gpointer     userdata)
{
  g_print ("callback invoked '%s' %p\n", string, userdata);
  friend_confirm_call ("foofoo");
}

static void notify (Friend *friend,
                    const gchar *message)
{
  gchar foo[512];               
  g_sprintf(foo, message, friend->name);

  moblin_add_notification (foo, G_CALLBACK (callback), friend);
}

#define SET_STATE(new_state)             \
   do {                                  \
     gchar foo[512];                     \
     g_sprintf(foo, "%s %s→%s",          \
           friend->name,                 \
           state_strings[friend->state], \
           state_strings[new_state]);    \
     friend->state=new_state;            \
   } while (0)

static gboolean friend_activity (gpointer data)
{
  Friend *friend = data;
  static gint transition;

  for (; friend->name != NULL; friend++)
    {

  switch (friend->state)
    {
      case AWAY:
        if (g_random_double_range (0.0, 1.0) < 0.05)
          {
            SET_STATE(AWAKE);
            notify(friend, "%s came online.");
	    transition = 5;
          }
        break;

      case AWAKE:
        if (g_random_double_range (0.0, 1.0) < 0.05)
          {
            SET_STATE(AWAY);
            notify(friend, "%s went offline.");
	    transition = 5;
            break;
          }
        if (g_random_double_range (0.0, 1.0) < 0.1)
          SET_STATE(CHATTY);
        break;

      case CHATTY:
        if (g_random_double_range (0.0, 1.0) < 0.1)
          {
            SET_STATE(AWAKE);
            break;
          }
        if (g_random_double_range (0.0, 1.0) < 0.1)
          {
            SET_STATE(PRODDING);
            notify(friend, "%s sent an SMS.");
	    transition = 5;
            break;
          }
        if (g_random_double_range (0.0, 1.0) < 0.1)
          {
            SET_STATE(CALLING);
            notify(friend, "☎  %s is calling.");
	    transition = 5;
            break;
          }
        break;

      case CALLING:
        if (g_random_double_range (0.0, 1.0) < 0.1)
          SET_STATE(CHATTY);
        if (g_random_double_range (0.0, 1.0) < 0.1)
          SET_STATE(AWAKE);
        break;

      case PRODDING:
        if (g_random_double_range (0.0, 1.0) < 0.1)
          SET_STATE(CHATTY);
        if (g_random_double_range (0.0, 1.0) < 0.1)
          SET_STATE(AWAKE);
        break;
    }
    }

  if (transition) {
	  transition--;
	  if (!transition)
		  tween (moblin_desktop()->notification_group, 500, 
			      "opacity", 0.0, NULL);
  }

  return TRUE;
}

void mock_friends (void)
{
      g_timeout_add(4000, friend_activity, friends);
}


typedef struct Scene
{
  ClutterActor     *root;
  ClutterScript    *script;
  ClutterTimeline  *appear;
  ClutterTimeline  *disappear;
} Scene;

Scene *
launch_script (const gchar *path)
{
  ClutterActor *modal_layer = moblin_desktop ()->modal;
  Scene *scene;
  GError *error = NULL;

  moblin_blur ();
  scene = g_new0 (Scene, 1);
 
  scene->script = clutter_script_new ();
  clutter_script_load_from_file (scene->script, path, &error);

  if (error)
    {
      g_print ("%s\n", error->message);
      g_clear_error (&error);
      return NULL;
    }
  clutter_script_connect_signals (scene->script, scene);
  scene->root = CLUTTER_ACTOR (clutter_script_get_object (scene->script, "root"));

  if (clutter_script_get_object (scene->script, "appear"))
    scene->appear = CLUTTER_TIMELINE (clutter_script_get_object (scene->script, "appear"));
  if (clutter_script_get_object (scene->script, "disappear"))
    scene->disappear = CLUTTER_TIMELINE (clutter_script_get_object (scene->script, "disappear"));

  clutter_group_add (CLUTTER_GROUP (modal_layer), scene->root);

  if (scene->appear)
    clutter_timeline_start (scene->appear);
  return scene;
}


void
launch_script_as_app (const gchar *path)
{
  Scene  *scene;
  GError *error = NULL;

  scene = g_new0 (Scene, 1);
 
  scene->script = clutter_script_new ();
  clutter_script_load_from_file (scene->script, path, &error);

  if (error)
    {
      g_print ("%s\n", error->message);
      g_clear_error (&error);
      return;
    }
  clutter_script_connect_signals (scene->script, scene);
  scene->root = CLUTTER_ACTOR (clutter_script_get_object (scene->script, "root"));

  moblin_desktop_activate_application (scene->root);

  if (clutter_script_get_object (scene->script, "appear"))
    scene->appear = CLUTTER_TIMELINE (clutter_script_get_object (scene->script, "appear"));
  if (clutter_script_get_object (scene->script, "disappear"))
    scene->disappear = CLUTTER_TIMELINE (clutter_script_get_object (scene->script, "disappear"));

  if (scene->appear)
    clutter_timeline_start (scene->appear);
}

void
friend_confirm_call (const gchar *friend)
{
  ClutterActor *name;
  Scene        *scene;
  scene = launch_script (JSON_ROOT "confirm-call.json");
  name = (void*)clutter_script_get_object (scene->script, "name");
  if (name)
    clutter_label_set_text (CLUTTER_LABEL (name), friend);

/*  g_print ("Launched app: %s\n", friend);*/

  tween (moblin_desktop()->notification_group, 1000, "opacity", 0.0, NULL);
}

gboolean
confirm_call_confirm_event (ClutterActor *foo,
                            ClutterEvent *event,
                            gpointer      data)
{
  Scene *scene = data;
  clutter_actor_hide    (scene->root);
  clutter_actor_destroy (scene->root); 
  moblin_unblur ();
  launch_script_as_app (JSON_ROOT "in-call.json");
  tween (moblin_desktop()->notification_group, 1000,
              "opacity", 0.0, NULL);
  moblin_fullscreen (TRUE);
  return FALSE;
}

gboolean
confirm_call_reject_event (ClutterActor *foo,
                           ClutterEvent *event,
                           gpointer      data)
{
  Scene *scene = data;

  clutter_actor_hide (scene->root);
  clutter_actor_destroy (scene->root);
  moblin_unblur ();
  tween (moblin_desktop()->notification_group, 1000,
              "opacity", 255.0, NULL);
  return FALSE;
}


gboolean in_call_end_call_event (ClutterActor *foo,
                                 ClutterEvent *event,
                                 gpointer      data)
{
  Scene *scene = data;

  clutter_actor_destroy (scene->root);
  g_free (scene);

  tween (moblin_desktop()->middle,
     500,
     "scale", 1.0,
     "x", 0.0,
     "y", 0.0,
     NULL);

  return FALSE;
}
