/* 2008 (c) Intel Corporation,  All rights reserved.
 *
 * Original author Øyvind Kolås <pippin@openedhand.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     - Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     - Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     - Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Intel Corporation ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL Intel Corporation BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <clutter/clutter.h>
#include "moblin-desktop.h"
#include "clutter-box2d.h"

static ClutterColor through_border_color = {0x00,0xff,0xff,0x22};
static ClutterColor puck_color           = {0x77,0x88,0xbb,0xaa};
static ClutterColor page_color           = {0x77,0x88,0xbb,0xff};
#define BORDER_WIDTH              4
#define PUCKSIZE                           32


static ClutterActor *add_dynamic_box (ClutterActor *box2d,
                                      gint          x,
                                      gint          y,
                                      gint          width,
                                      gint          height,
                                      ClutterColor *color);
static ClutterActor * add_static_box (ClutterActor *box2d,
                                      gint          x,
                                      gint          y,
                                      gint          width,
                                      gint          height,
                                      ClutterColor *color);
static void add_static_boundary      (ClutterActor *box2d,
                                      gint          x,
                                      gint          y,
                                      gint          width,
                                      gint          height,
                                      ClutterColor *color);
static void zero_gravity             (ClutterActor *box2d)
{
  ClutterVertex gravity = {0,0};
  g_object_set (box2d, "gravity", &gravity, NULL);
}



static ClutterActor * add_static_box (ClutterActor *box2d,
                                      gint          x,
                                      gint          y,
                                      gint          width,
                                      gint          height,
                                      ClutterColor *color)
{ 
   ClutterActor *border;
   border = clutter_rectangle_new ();
   clutter_actor_set_size (border, width, height);
   clutter_actor_set_position (border, x, y);
   clutter_group_add (CLUTTER_GROUP (box2d), border);

   clutter_rectangle_set_color (CLUTTER_RECTANGLE (border), color);
   clutter_container_child_set (CLUTTER_CONTAINER (box2d), border,
                                "mode", CLUTTER_BOX2D_STATIC,
                                NULL);
   clutter_actor_set_opacity (border, 255);
   return border;
}

static ClutterActor *add_dynamic_box (ClutterActor *box2d,
                                      gint          x,
                                      gint          y,
                                      gint          width,
                                      gint          height,
                                      ClutterColor *color)
{ 
   ClutterActor *border;
   border = clutter_rectangle_new ();
   clutter_actor_set_size (border, width, height);
   clutter_actor_set_position (border, x, y);
   clutter_group_add (CLUTTER_GROUP (box2d), border);

   clutter_rectangle_set_color (CLUTTER_RECTANGLE (border), color);
   clutter_container_child_set (CLUTTER_CONTAINER (box2d), border,
                                "manipulatable", TRUE,  /* user can interact with item */
                                "mode", CLUTTER_BOX2D_DYNAMIC,
                                NULL);
   return border;
}

static void add_static_boundary (ClutterActor *box2d,
                                 gint          x,
                                 gint          y,
                                 gint          width,
                                 gint          height,
                                 ClutterColor *color)
{
  ClutterActor *top,*bottom,*left,*right;
  gint pad = BORDER_WIDTH;

  /* zero gravity */

  top = add_static_box    (box2d, x-pad,   y-pad, width + pad*2, pad, color);
  bottom = add_static_box (box2d, x-pad,   y+height, width + pad*2, pad, color);
  left = add_static_box   (box2d, x-pad,   y,     pad, height, color);
  right = add_static_box  (box2d, x+width, y,     pad, height, color);
}

static void track_hslider (ClutterActor *actor,
                           GParamSpec   *pspec,
                           gpointer      cb)
{
  gpointer cb_data = g_object_get_data (G_OBJECT (actor), "cb-data");
  gint x;
  gdouble width;

  void (*value_changed) (ClutterActor *slider,
                         gdouble       rel_val,
                         gpointer      user_data)= cb;

  x = clutter_actor_get_x (actor);
  width = clutter_actor_get_width (clutter_actor_get_parent (actor));

  width -= BORDER_WIDTH;
  width -= PUCKSIZE;

  if (x<0)
    x=0;

  value_changed (actor, x/width, cb_data);
}

#if 0
static void track_vslider (ClutterActor *actor,
                           GParamSpec   *pspec,
                           gpointer      cb)
{
  gpointer cb_data = g_object_get_data (G_OBJECT (actor), "cb-data");
  gint y;
  gdouble height;

  void (*value_changed) (ClutterActor *slider,
                         gdouble       rel_val,
                         gpointer      user_data)= cb;

  y = clutter_actor_get_y (actor);
  height = clutter_actor_get_height (clutter_actor_get_parent (actor));

  height -= BORDER_WIDTH;
  height -= PUCKSIZE;

  if (y<0)
    y=0;

  value_changed (actor, y/height, cb_data);
}
#endif

static void add_hslider (ClutterActor *group,
                         gint          x,
                         gint          y,
                         gint          width,
                         GCallback     callback,
                         gpointer      user_data)
{
  ClutterActor *box2d = clutter_box2d_new ();
  ClutterActor *puck;

  zero_gravity (box2d);
  clutter_box2d_set_simulating (CLUTTER_BOX2D (box2d), TRUE);

  add_static_boundary (box2d, 0, 0, width, PUCKSIZE, &through_border_color);
  puck = add_dynamic_box (box2d, 3, 1, PUCKSIZE, PUCKSIZE, &puck_color);

  clutter_group_add (group, box2d);

  clutter_actor_set_position (box2d, x, y);
 
  if (callback)
    { 
      g_signal_connect (G_OBJECT (puck), "notify::allocation", G_CALLBACK (track_hslider), callback);
      g_object_set_data (G_OBJECT (puck), "cb-data", user_data);
    }
}

#if 0
static void add_vslider (ClutterActor *group,
                         gint          x,
                         gint          y,
                         gint          height,
                         GCallback     callback,
                         gpointer      user_data)
{
  ClutterActor *box2d = clutter_box2d_new ();
  ClutterActor *puck;

  zero_gravity (box2d);
  clutter_box2d_set_simulating (CLUTTER_BOX2D (box2d), TRUE);
  add_static_boundary (box2d, 0, 0, PUCKSIZE, height, &through_border_color);
  puck = add_dynamic_box (box2d, 1, 3, PUCKSIZE, PUCKSIZE, &puck_color); 

  clutter_group_add (group, box2d);
  clutter_actor_set_position (box2d, x, y);

  if (callback)
    { 
      g_signal_connect (G_OBJECT (puck), "notify::allocation",
                        G_CALLBACK (track_vslider), callback);
      g_object_set_data (G_OBJECT (puck), "cb-data", user_data);
    }
}
#endif

static void red_changed (ClutterActor *slider,
                         gdouble       rel_val,
                         gpointer      user_data)
{
  ClutterColor  color;
  ClutterRectangle *rectangle = CLUTTER_RECTANGLE (user_data);

  clutter_rectangle_get_color (rectangle, &color);
  color.red = rel_val * 255;
  clutter_rectangle_set_color (rectangle, &color);
}

static void green_changed (ClutterActor *slider,
                         gdouble       rel_val,
                         gpointer      user_data)
{
  ClutterColor  color;
  ClutterRectangle *rectangle = CLUTTER_RECTANGLE (user_data);

  clutter_rectangle_get_color (rectangle, &color);
  color.green = rel_val * 255;
  clutter_rectangle_set_color (rectangle, &color);
}


static void blue_changed (ClutterActor *slider,
                          gdouble       rel_val,
                          gpointer      user_data)
{
  ClutterColor  color;
  ClutterRectangle *rectangle = CLUTTER_RECTANGLE (user_data);

  clutter_rectangle_get_color (rectangle, &color);
  color.blue = rel_val * 255;
  clutter_rectangle_set_color (rectangle, &color);
}


static void alpha_changed (ClutterActor *slider,
                         gdouble       rel_val,
                         gpointer      user_data)
{
  ClutterColor  color;
  ClutterRectangle *rectangle = CLUTTER_RECTANGLE (user_data);

  clutter_rectangle_get_color (rectangle, &color);
  color.alpha = rel_val * 255;
  clutter_rectangle_set_color (rectangle, &color);
}

#if 0
static void value_changedv (ClutterActor *slider,
                           gdouble       rel_val,
                           gpointer      user_data)
{
  g_print ("(%p changed to  d %f (ud=%p)\n", slider, rel_val, user_data);
}
#endif

ClutterActor *settings_new (const gchar *path)
{
  ClutterActor *box2d;
  ClutterActor *background;
  ClutterActor *page;
  ClutterActor *color = NULL;
  ClutterActor *group = clutter_group_new ();

  gint x, y, width, height;

  box2d = clutter_box2d_new ();
  zero_gravity (box2d);
  background = clutter_texture_new_from_file (ASSET_ROOT "background.png", NULL);

  clutter_actor_set_width (background, clutter_actor_get_width (moblin_desktop()->stage));

  clutter_group_add (CLUTTER_GROUP (group), background);
  
  color = clutter_rectangle_new ();
  clutter_group_add (CLUTTER_GROUP (group), color);
  clutter_actor_set_size (color, 100, 100);
  clutter_actor_set_position (color, 350, 30);

  add_hslider (group, 10, 10 + 50 * 0, 300, G_CALLBACK (red_changed), color);
  add_hslider (group, 10, 10 + 50 * 1, 300, G_CALLBACK (green_changed), color);
  add_hslider (group, 10, 10 + 50 * 2, 300, G_CALLBACK (blue_changed), color);
  add_hslider (group, 10, 10 + 50 * 3, 300, G_CALLBACK (alpha_changed), color);

  /*add_vslider (group, 300 + 50 * 1, 10, 300, G_CALLBACK (value_changedv), NULL);
  add_vslider (group, 300 + 50 * 2, 10, 300, NULL, NULL);
  add_vslider (group, 300 + 50 * 3, 10, 300, NULL, NULL);*/


  actor_get_extent (group, &x, &y, &width, &height);
  add_static_boundary (box2d, x, y, width, 800, &through_border_color);
  page = add_dynamic_box (box2d, x, y+3, width, height, &page_color);

  clutter_group_add (CLUTTER_GROUP (box2d), group);

  clutter_box2d_actor_track (group, page, CLUTTER_BOX2D_TRACK_POSITION);

  clutter_box2d_set_simulating (CLUTTER_BOX2D (box2d), TRUE);
                                //"manipulatable", TRUE,  /* user can interact with item */

  return box2d;
}
