/* 2008 (c) Intel Corporation,  All rights reserved.
 *
 * Original author Øyvind Kolås <pippin@openedhand.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     - Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     - Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     - Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Intel Corporation ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL Intel Corporation BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <clutter/clutter.h>
#include "moblin-desktop.h"
#include "clutter-box2d.h"

static gdouble orig_x;
static gdouble orig_y;
static gdouble orig_rot;
static gdouble orig_scale;

#define SHADOW_OFFSET_X 20
#define SHADOW_OFFSET_Y 20

static gboolean nuke_it (ClutterActor *item)
{
  clutter_actor_destroy (item);
  return FALSE;
}

static gboolean nuke_clone_event (ClutterActor *item,
                                  ClutterEvent *event,
                                  gpointer      userdata)
{
  if (g_object_get_data (G_OBJECT (item), "kthxbye"))
    return FALSE;
  /* a check hack for marking objects we're going to destroy */
  g_object_set_data (G_OBJECT (item), "kthxbye", (gpointer)0xff);
  tween (item, 300, "scale", orig_scale,
              "x", orig_x,
              "y", orig_y,
              "rotation", orig_rot,
              "opacity", 0.0,
              NULL);
  g_timeout_add (800, (gpointer)nuke_it, item);

  return TRUE;
}

static gboolean item_release_event (ClutterActor *item,
                                    ClutterEvent *event,
                                    gpointer      userdata)
{
  if (event->button.click_count == 2)
    {
      ClutterActor *clone = NULL;
      ClutterActor *orig_texture;
      ClutterActor *group = clutter_actor_get_parent (item);
      gint x, y;
      gint w, h;
      gdouble rot;
      gdouble scale;

      { /* fetch the actual texture from inside the group consisting of
         * background+texture, this bit is perhaps a little too hard-coded
         */
        GList *children = clutter_container_get_children (CLUTTER_CONTAINER (item));
        GList *iter = children;

        if (iter)
          {
            iter = iter->next;
            g_assert (iter);
            orig_texture = iter->data;
          }
        g_list_free (children);
      }

      clone = clutter_clone_texture_new (CLUTTER_TEXTURE (orig_texture));

      clutter_group_add (CLUTTER_GROUP (group), clone);
      clutter_actor_get_position (item, &x, &y);
      rot = clutter_actor_get_rotation (item, CLUTTER_Z_AXIS, 0,0,0);
      clutter_actor_set_position (clone, x, y);
      clutter_actor_set_rotation (clone, CLUTTER_Z_AXIS, rot, 0,0,0);

      orig_x = x;
      orig_y = y;
      orig_rot = rot;
      orig_scale = 1.0;

      w = clutter_actor_get_width (clone);
      h = clutter_actor_get_height (clone);

      if (w>h)
        {
          scale = 1.0 * moblin_desktop_width ()/w;
          if (scale * h > moblin_desktop_height ())
            scale = 1.0* moblin_desktop_height ()/h;
        }
      else
        {
          scale = 1.0* moblin_desktop_height ()/h;
          if (scale * w > moblin_desktop_width ())
            scale = 1.0* moblin_desktop_width ()/w;
        }


      /* tween to fullscreen (or at least full application screen, the
       * application might not be running fullscreen)
       */
      tween (clone, 500,
                  "scale", scale,
                  "x", (moblin_desktop_width () -(w*scale))/2.0,
                  "y", (moblin_desktop_height () -(h*scale))/2.0,
                  "rotation", 0.0,
                  NULL);

      clutter_actor_set_reactive (clone, TRUE);
      g_signal_connect (clone, "button-press-event", G_CALLBACK (nuke_clone_event),
                        NULL);

      return FALSE;
    }
  return FALSE;
}

static void add_static_box (ClutterActor *box2d,
                            gint          x,
                            gint          y,
                            gint          width,
                            gint          height)
{ /* add wborders */
   ClutterActor *border;
   border = clutter_rectangle_new ();
   clutter_actor_set_size (border, width, height);
   clutter_actor_set_position (border, x, y);
   clutter_group_add (CLUTTER_GROUP (box2d), border);

   clutter_container_child_set (CLUTTER_CONTAINER (box2d), border,
                                "mode", CLUTTER_BOX2D_STATIC,
                                NULL);
   clutter_actor_set_opacity (border, 0);
}

static void frame_stage (ClutterActor *box2d)
{
  gint stage_width = moblin_desktop_width ();
  gint stage_height = moblin_desktop_height ();

  add_static_box (box2d, 0,              3,    stage_width, 10);
  add_static_box (box2d, 0,   stage_height-10, stage_width, 10);
  add_static_box (box2d, 0,              3,    10,          stage_height);
  add_static_box (box2d, stage_width-10, 3,    10,          stage_height);
}

ClutterActor *table_new (const gchar *path)
{
  gint         i;
  gchar      **dir_entries;
  ClutterColor shadow_color = {0,0,0,0x44};
  gint         count;

  ClutterActor *box2d    = clutter_box2d_new ();
  ClutterVertex gravity  = {0,0};

  if (path == NULL)
    path = ASSET_ROOT "photos/";

  dir_entries = get_dir_entries (path);
  count = pasp_count (dir_entries);


  /* zero gravity */
  g_object_set (box2d, "gravity", &gravity, NULL);

  frame_stage (box2d);

  for (i=0; i<12 ; i++)
    {
      gint w, h;
      ClutterActor *group;
      ClutterActor *image;
      ClutterActor *rectangle;
      ClutterActor *shadow;

      group = clutter_group_new ();

      image = g_object_new (CLUTTER_TYPE_TEXTURE,
        "filename", dir_entries[g_random_int_range (0, count)],
        "filter-quality", CLUTTER_TEXTURE_QUALITY_HIGH,
        NULL);
/*        
        clutter_texture_new_from_file (
                   dir_entries[ g_random_int_range (0, count) ], NULL);*/

      w = clutter_actor_get_width (image) + 2;
      h = clutter_actor_get_height (image) + 2;

      shadow = clutter_rectangle_new ();
      rectangle = clutter_rectangle_new ();
      clutter_actor_set_size (rectangle, w, h);
      clutter_actor_set_size (shadow, w, h);
      clutter_actor_set_position (image, 1, 1);
      clutter_actor_set_opacity (rectangle, 255*0.75);
      clutter_rectangle_set_color (CLUTTER_RECTANGLE (shadow), &shadow_color);

      clutter_container_add (CLUTTER_CONTAINER (group), rectangle, image, NULL);

      /* add image centered, with a slight jitter to more easily let box2d
       * resolve the positioning at run time
       */
      clutter_actor_set_position (group,
                                  moblin_desktop_width ()/2 + g_random_int_range (-(moblin_desktop_width () / 4), (moblin_desktop_width () / 4)),
                                  moblin_desktop_height ()/2 + g_random_int_range (-(moblin_desktop_height () / 4), (moblin_desktop_height () / 4)));

      clutter_actor_set_position (shadow,
                                  clutter_actor_get_x (group),
                                  clutter_actor_get_y (group));

      clutter_group_add (CLUTTER_GROUP (box2d), shadow);
      clutter_group_add (CLUTTER_GROUP (box2d), group);
      clutter_actor_set_depth (shadow, -50);

      /* lower the shadow to the bottom, so it ends up beneath already added images in
       * the box2d container
       */
      if(0)clutter_actor_lower_bottom (shadow);

      /* has to be done after adding them both to their containers */
      clutter_box2d_actor_track (shadow, group, CLUTTER_BOX2D_TRACK_ALL);

      clutter_container_child_set (CLUTTER_CONTAINER (box2d), group,
                                   "manipulatable", TRUE,  /* user can interact with item */
                                   "mode",          CLUTTER_BOX2D_DYNAMIC,
                                   NULL);
     
      clutter_box2d_set_simulating (CLUTTER_BOX2D (box2d), TRUE);

      g_signal_connect (group, "button-press-event",
                        G_CALLBACK (item_release_event), NULL);
    }
  g_free (dir_entries);
  return box2d;
}
