/* 2008 (c) Øyvind Kolås <pippin@openedhand.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     - Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     - Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     - Neither the name of Øyvind Kolås nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Øyvind Kolås ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL Øyvind Kolås BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdlib.h>
#include <stdarg.h>
#include <clutter/clutter.h>

/* comment out to enable experimental path building mode */
#define ALWAYS_REPLACE     

/* The tweening info is stored in a list stored under the
 * userdata tween on the actor. As long as there are tweens
 * existing a timeline is attached that consumes the delay
 * until deadline passes, when deadline passses the property
 * should have arrived at its destination.
 */
typedef struct Tween {
  ClutterActor *actor;         /* The actor this */
  gchar        *property;      /* name of property */
  ClutterUnit   start;         /* initial value for named property */
  ClutterUnit   target;        /* destination value for named property */
  ClutterAlphaFunc alpha_func;

  gint          deadline;      /* when should the tween be complete */
  guint         elapsed;       /* elapsed + deadline == duration */

  struct Tween *next;          /* further tweens of this property to be
                                  commenced after the current one is done
                                  (stored with deadlines running after the
                                  next expiring tween for this property is
                                  finished).*/
} Tween;

static GSList          *tweens = NULL;
static ClutterTimeline *tweener = NULL;
static ClutterTimeline *zombie  = NULL;
static ClutterAlpha    *zombie_alpha = NULL;

static void actor_dies (gpointer data,
                        GObject *where_the_object_was)
{
  Tween *id = data;
  id->actor = NULL;
  tweens = g_slist_remove (tweens, id);
  g_free (id);
}

static void
on_new_frame (ClutterTimeline *timeline, 
	      gint             frame_num, 
	      gpointer         data)
{
  gint   msecs;
  GSList *iter;
  GSList *prev = NULL;

  clutter_timeline_get_delta (timeline, (void*)&msecs);
 
  if (msecs==0)
    msecs=1;

#define VALUE_APPROACH_TARGET_LINEAR() \
          value = CLUTTER_UNITS_FROM_FLOAT(                 \
              CLUTTER_UNITS_TO_FLOAT (value) +              \
             (CLUTTER_UNITS_TO_FLOAT (target) -             \
              CLUTTER_UNITS_TO_FLOAT (value)) / (tween->deadline/msecs) \
           );                                               \
          if (tween->deadline-msecs-2 <= 0)                 \
            {                                               \
              value = target;                               \
              if (tween->next)                              \
                {                                           \
                  iter->data = tween->next;                 \
                }                                           \
              else                                          \
                {                                           \
                  g_object_weak_unref (G_OBJECT (tween->actor), actor_dies, tween);\
                  tweens = g_slist_remove (tweens, tween);  \
                }                                           \
            }                                               \
          else                                              \
            {                                               \
              tween->deadline -= msecs;                       \
              tween->elapsed += msecs;                       \
              prev = iter;                                  \
            }

#define VALUE_APPROACH_TARGET_PORTION() \
          value = CLUTTER_UNITS_FROM_FLOAT(                \
             (0.9 * CLUTTER_UNITS_TO_FLOAT(value) +        \
              0.1 * CLUTTER_UNITS_TO_FLOAT (target))       \
           );                                               \
          if (abs(value-target)<255)\
            {                                               \
              tweens = g_slist_remove (tweens, tween);      \
              g_free (tween);                               \
            }                                               \
          else                                              \
            {                                               \
              prev = iter;                                  \
            }

/* these should be converted to functions, or perhaps be made
 * possible to choose through the faked "alpha" property.
 */

#define VALUE_APPROACH_TARGET_ALPHA() \
          if (tween->deadline-msecs-2 <= 0)                 \
            {                                               \
              value = target;                               \
              if (tween->next)                              \
                {                                           \
                  iter->data = tween->next;                 \
                }                                           \
              else                                          \
                {                                           \
                  tweens = g_slist_remove (tweens, tween);  \
                }                                           \
            }                                               \
          else                                              \
            {                                               \
              tween->deadline -= msecs;                       \
              tween->elapsed += msecs;                       \
              prev = iter;                                  \
            }                                               \
          value =                      \
              CLUTTER_FIXED_MUL(65535-alpha, start)+ \
              CLUTTER_FIXED_MUL(alpha, target);

/*#define VALUE_APPROACH_TARGET() VALUE_APPROACH_TARGET_ALPHA()*/
#define VALUE_APPROACH_TARGET() VALUE_APPROACH_TARGET_LINEAR()

  for (iter=tweens; iter; iter = iter->next)
    {
      Tween        *tween  = iter->data;
      ClutterActor *actor  = tween->actor;
      ClutterUnit   start;
      ClutterUnit   target;
      guint32       alpha;

      start = tween->start;
      target = tween->target;

      clutter_timeline_stop (zombie);
      clutter_timeline_advance (zombie, 256.0*tween->elapsed/(tween->deadline+tween->elapsed));
      alpha = tween->alpha_func (zombie_alpha, NULL);

      if (g_str_equal (tween->property, "depth"))
        {
          ClutterUnit value = clutter_actor_get_depthu (actor);
          VALUE_APPROACH_TARGET();
          clutter_actor_set_depthu (actor, value);
        }
      else if (g_str_equal (tween->property, "x"))
        {
          ClutterUnit value = clutter_actor_get_xu (actor);
          VALUE_APPROACH_TARGET();
          clutter_actor_set_xu (actor, value);
        }
      else if (g_str_equal (tween->property, "y"))
        {
          ClutterUnit value = clutter_actor_get_yu (actor);
          VALUE_APPROACH_TARGET();
          clutter_actor_set_yu (actor, value);
        }
      else if (g_str_equal (tween->property, "scale"))
        {
          ClutterUnit value;
          clutter_actor_get_scalex (actor, &value, NULL);
          VALUE_APPROACH_TARGET();
          clutter_actor_set_scalex (actor, value, value);
        }
      else if (g_str_equal (tween->property, "rotation"))
        {
          ClutterUnit value;
          gint x, y, z;
          value = clutter_actor_get_rotationx (actor, CLUTTER_Z_AXIS, &x, &y, &z);
          VALUE_APPROACH_TARGET();
          clutter_actor_set_rotationx (actor, CLUTTER_Z_AXIS, value, x, y, z);
        }
      else if (g_str_equal (tween->property, "opacity"))
        {
          ClutterUnit value = CLUTTER_UNITS_FROM_INT (
                         clutter_actor_get_opacity (actor));
          VALUE_APPROACH_TARGET();
          clutter_actor_set_opacity (actor, CLUTTER_UNITS_TO_INT (value));
        }
      else 
        {
          g_warning ("unhandled tween property %s", tween->property);
        }
      clutter_actor_queue_redraw (actor);
    }
}


static void tween_one_x (ClutterActor    *actor,
                         guint            deadline,
                         ClutterAlphaFunc alpha_func,
                         const gchar     *property,
                         ClutterUnit      target)
{
  Tween  *id;
 
  id = g_new0(Tween, 1);

  id->actor = actor;
  id->elapsed = 0;
  id->deadline = deadline;
  id->target = target;
  id->property = g_strdup (property);
  id->next = NULL;
  id->alpha_func = alpha_func;

  g_object_weak_ref (G_OBJECT (actor), actor_dies, id);

  id->target = target;

  if (g_str_equal (id->property, "depth"))
    {
      id->start = clutter_actor_get_depthu (actor);
    }
  else if (g_str_equal (id->property, "x"))
    {
      id->start = clutter_actor_get_xu (actor);
    }
  else if (g_str_equal (id->property, "y"))
    {
      id->start = clutter_actor_get_yu (actor);
    }
  else if (g_str_equal (id->property, "scale"))
    {
      ClutterUnit value;
      clutter_actor_get_scalex (actor, &value, NULL);
      id->start = value;
    }
  else if (g_str_equal (id->property, "rotation"))
    {
      gint x, y, z;
      id->start = clutter_actor_get_rotationx (actor, CLUTTER_Z_AXIS, &x, &y, &z);
    }
  else if (g_str_equal (id->property, "opacity"))
    {
      id->start = CLUTTER_UNITS_FROM_INT (
                   clutter_actor_get_opacity (actor));
    }
  else 
    {
      g_warning ("unhandled tween property %s", id->property);
    }

  /* Iterate through all pending tweens */
  {
    GSList *iter;
    for (iter= tweens; iter; iter=iter->next)
      {
        Tween *old = iter->data;

        if (old->actor == actor &&
            g_str_equal (old->property, property))
          /* we've found a pre-existing tween for this same property */
          {
#ifdef ALWAYS_REPLACE
                tweens = g_slist_remove (tweens, old);
                g_object_weak_unref (G_OBJECT (old->actor), actor_dies, old);
                g_free (old);
#else
            /* if the new proposed deadline is earlier than the one already
             * set, we replace it
             */
            if (deadline <= old->deadline)
              {
                /* FIXME: nuke all tweens for a prop */
                tweens = g_slist_remove (tweens, old);
                g_object_weak_unref (G_OBJECT (old->actor), actor_dies, old);
                g_free (old);
              }
            else
              {
                /* otherwise we append ourselves as a next tweening state,
                 * (this should be done recursivly but isn't,..
                 */

                while (old->next &&
                       deadline > old->deadline)
                  {
                    deadline -= old->deadline;
                    old = old->next;
                  }

                id->next = old->next;
                old->next = id;
                return;
              }
            break;
#endif
          }
      }
  }

  /* FIXME: add a weak reference on the actor */

  tweens = g_slist_prepend (tweens, id);

  if (tweener == NULL)
    {
      tweener = clutter_timeline_new (100, 25);
      g_object_set(tweener, "loop", TRUE, NULL);
      g_signal_connect (tweener, "new-frame", G_CALLBACK (on_new_frame), NULL);
      clutter_timeline_start (tweener);

      zombie = clutter_timeline_new (255, 25);
      zombie_alpha = clutter_alpha_new ();
      clutter_alpha_set_timeline (zombie_alpha, zombie);
    }
}

static void tween_one (ClutterActor    *actor,
                       guint            deadline,
                       ClutterAlphaFunc alpha_func,
                       const gchar     *property,
                       gdouble          value)
{
  tween_one_x (actor, deadline, alpha_func, 
               property, CLUTTER_UNITS_FROM_FLOAT (value));
}

void tween (ClutterActor *actor,
                 guint         deadline,
                 const gchar  *first_property,
                 ...)
{
  const gchar *property_name;
  va_list var_args;
  va_start (var_args, first_property);
  property_name = first_property;
  ClutterAlphaFunc alpha_func = CLUTTER_ALPHA_RAMP_INC;
  
  while (property_name)
    {
      gdouble value;
      if (g_str_equal (property_name, "alpha"))
       {
         alpha_func = va_arg (var_args, ClutterAlphaFunc);
         property_name = va_arg (var_args, gchar *);
         continue; /* we swallowed this and updated our state*/
       }
      value = va_arg (var_args, gdouble);
      tween_one (actor, deadline, alpha_func, property_name, value);

      property_name = va_arg (var_args, gchar *);
    }
  if(0)g_print ("\n");
  va_end (var_args);
}
