/* 2008 (c) Intel Corporation,  All rights reserved.
 *
 * Original author Øyvind Kolås <pippin@openedhand.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     - Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     - Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     - Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Intel Corporation ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL Intel Corporation BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <glib.h>
#include <glib/gprintf.h>
#include <string.h>
#include <clutter/clutter.h>

/* This function returns a argv* like array in a single allocation (to be freed
 * with a single g_free ()
 *
 * The path passed in should not end with a '/' for the created composite paths
 * to work correctly.
 */
gchar **
get_dir_entries (const gchar *path)
{
  gchar   tempbuf[512];
  gchar **argv         = NULL;
  gchar  *string       = NULL;
  gint    items;
  gint    total_length = 0;
  gint    offset       = 0;

  /* first a pass to compute our total memory requirement */ 
  items = 0; 
  {
    GDir   *dir;
    const gchar  *fname;
    dir = g_dir_open (path, 0, NULL);
    if (!dir)
      {
        argv = g_malloc0 (sizeof (gchar *));
        return argv;
      }
    while ((fname = g_dir_read_name (dir)))
      {
        g_sprintf (tempbuf, "%s/%s", path, fname);

        if (g_file_test (tempbuf, G_FILE_TEST_IS_REGULAR))
          {
            total_length += strlen (tempbuf) + 1;
            items ++;
          }
      }
    g_dir_close (dir);
  }

  argv = g_malloc0 ( sizeof (gchar *) * (items+1) +
                    sizeof (gchar) * (total_length));
   
  string = (gchar*)argv;

  offset = sizeof (gchar *) * (items+1);
  items = 0;

  /* then a second pass populating our allocated memory */
  {
    GDir   *dir;
    const gchar  *fname;
    dir = g_dir_open (path, 0, NULL);
    if (!dir)
      return NULL;
    while ((fname = g_dir_read_name (dir)))
      {
        g_sprintf (tempbuf, "%s/%s", path, fname);

        if (g_file_test (tempbuf, G_FILE_TEST_IS_REGULAR))
          {
            argv[items] = &string[offset];
            memcpy (&string[offset], tempbuf, strlen (tempbuf)+1);
            offset += strlen (tempbuf) + 1;
            items ++;
          }
      }
    g_dir_close (dir);
  }
  return argv;
}

gint
pasp_count (gchar **argv)
{
  gchar **iter;
  gint    count;

  for (iter=argv, count=0; *iter; iter++,count++);
  return count;
}


/* retrieve the position of an actor relative to any
 * anchestor actor */
void
actor_get_rel_pos (ClutterActor *actor,
                   ClutterActor *anchestor,
                   gint         *x_ret,
                   gint         *y_ret)
{
  ClutterActor *iter;
  gint          x, y;

  x = 0;
  y = 0;

  for (iter = actor; iter && iter != anchestor;
       iter = clutter_actor_get_parent (iter))
    {
      x += clutter_actor_get_x (iter);
      y += clutter_actor_get_y (iter);
    }
  if (x_ret)
    *x_ret = x;
  if (y_ret)
    *y_ret = y;
}


static void
expand_extent (ClutterActor *top_level,
               ClutterActor *actor,
               gint         *x0,
               gint         *y0,
               gint         *x1,
               gint         *y1)
{
  if (CLUTTER_IS_ACTOR (actor))
    {
      gint x, y;

      actor_get_rel_pos (actor, top_level, &x, &y);

      if (x < *x0)
        *x0 = x;
      if (y < *y0)
        *y0 = y;
      if ((gint)(x + clutter_actor_get_width (actor)) > *x1)
        *x1 = x + (gint)clutter_actor_get_width (actor);
      if ((gint)(y + clutter_actor_get_height (actor)) > *y1)
        *y1 = y + (gint)clutter_actor_get_height (actor);

      /* recurse into children
       */
      if (CLUTTER_IS_CONTAINER (actor))
        {
          GList *children;
          GList *iter;

          children = clutter_container_get_children (CLUTTER_CONTAINER (actor));

          for (iter = children; iter; iter = g_list_next (iter))
            {
              expand_extent (top_level, CLUTTER_ACTOR (iter->data),
                             x0, y0, x1, y1);
            }

          g_list_free (children);
        }
    }
}

void
actor_get_extent (ClutterActor *actor,
                  gint         *x0,
                  gint         *y0,
                  gint         *x1,
                  gint         *y1)
{
  gint int_x0 = 65530;
  gint int_y0 = 65530;
  gint int_y1 = -65530;
  gint int_x1 = -65530;

  expand_extent (actor, actor,
                 &int_x0, &int_y0, &int_x1, &int_y1);
  if (x0)
    *x0 = int_x0;
  if (x1)
    *x1 = int_x1 - int_x0;
  if (y0)
    *y0 = int_y0;
  if (y1)
    *y1 = int_y1 - int_y0;
}
